import pytest
from balsam.selection import from_dict, to_dict
from balsam.selection import Contains, In, And, Or, SQL, IContains


@pytest.mark.parametrize(
    'selection',
    [(Contains('x', 'y')), (In('x', [1, 2, 3])), (And(In('x', 'y'), Contains('u', 'v'))),
     (Or(In('x', 'y'), Contains('u', 'v'))),
     (And(Or(In('x', 'y'), Contains('u', 'v')), And(In('x', 'y'), Contains('u', 'v')))),
     (SQL('"name" in (1, 2, 3)')), (IContains('x', 'y'))])
def test_to_dict_then_from_dict_is_id(selection):
    assert from_dict(to_dict(selection)) == selection


@pytest.mark.parametrize('obj,expected', [({
    'operation': 'in',
    'left': 'campaignId',
    'right': [1, 2, 3]
}, In('campaignId', [1, 2, 3]))])
def test_from_dict_conversion_directl(obj, expected):
    assert from_dict(obj) == expected


@pytest.mark.parametrize('obj', [({'operation': 'in'}, 'hello')])
def test_from_dict_throws_value_error_on_failure(obj):
    with pytest.raises(ValueError):
        from_dict(obj)



# pylint: disable=unused-wildcard-import, wildcard-import
from .base import *
from .validators import *

__all__ = [
    "check", "data_frame", "integer", "list_of", "string", "floating", "boolean",
    "array", "dict_of", "ValidationError", "selection"
]



import pytest

from balsam.selection import List, In, Contains, And, Or, SQL, IContains


class TestSelectionToSql:
    @pytest.mark.parametrize('expr,sql',
                             [(List(1, 2, 3), '(1,2,3)'),
                              (In('helloId', [1, 2, 3]), '"helloId" in (1,2,3)'),
                              (Contains('name', 'sub'), '"name" like \'%%sub%%\''),
                              (And(In('id', [1, 2, 3]), In('pk', [4, 5, 6])),
                               '("id" in (1,2,3)) and ("pk" in (4,5,6))'),
                              (Or(In('id', [1, 2, 3]), In('pk', [4, 5, 6])),
                               '("id" in (1,2,3)) or ("pk" in (4,5,6))'),
                              (SQL('\"hello\" in (1, 2, 3)'), '\"hello\" in (1, 2, 3)'),
                              (IContains("name", "sub"), '"name" ilike \'%%sub%%\'')])
    def test_to_sql(self, expr, sql):
        assert expr.to_sql() == sql



# IMPROVEMENT: This code is becoming a bit messy and hard to follow, due to some
# distinctions that are made between output and input (settings, and inputs) and
# the type of the arguments (mainly data frame versus non-data frame). We might
# need to rethink this a bit in the future.
import datetime as dt
from argparse import ArgumentParser

from balsam import selection
from balsam.store import DataFrameRef, JsonRef


def run_as_cli_program(task, args=None, store=None):
    """Runs the task as a command line program.

    :param task: The task to execute, should be an instance of
        :class:`~AbstractTask`.
    :param args: The list of command line arguments.  Ommit to use the actual
        command line arguments.
    :param store: The store to use to resolve references, such as
        :class:`~DataFrameRef` instances.

    :returns: The output of the task, the form of which is specified in the
              output schema of the task.
    """

    parser = create_argument_parser(task, no_output=False)
    if args is None:
        values = parser.parse_args()
    else:
        values = parser.parse_args(args)
    settings = _extract_keys(task.settings_schema.keys(), values)
    inputs = _extract_keys(task.input_schema.keys(), values)
    if store is not None:
        inputs = _resolve_inputs(store, inputs)
    output = task(settings, inputs)
    for key in task.output_schema:
        location = getattr(values, key)
        if hasattr(location, 'write'):
            location.write(output[key], store)
    return output


def _extract_keys(keys, values):
    result = {}
    for key in keys:
        result[key] = getattr(values, key)
    return result


def _resolve_inputs(store, inputs):
    result = {}
    for (key, value) in inputs.items():
        if hasattr(value, 'read'):
            result[key] = value.read(store)
        else:
            result[key] = value
    return result


class TaskArgumentParser(ArgumentParser):
    def __init__(self, task, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._task = task
        self._values = None

    def parse_args(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        if not self._values:
            self._values = super().parse_args(*args, **kwargs)
        return self._values

    @property
    def settings(self):
        self.parse_args()
        return _extract_keys(self._task.settings_schema.keys(), self._values)


def create_argument_parser(task, *args, no_output=False, **kwargs):
    """Creates an argument parser for the task.

    Currently quite limited in what kind of task schemas it accepts.

    :param task: A task, i.e., something that complies with the :class:`~Task`
        interface.
    :param no_output: Do not parse output parameters.
    :param args, kwargs: They are passed directly to the constructor of the
        :class:`~ArgumentParser`.

    :returns: A Python argument parser.
    """
    parser = TaskArgumentParser(task, *args, **kwargs)
    _add_schema_to_group(parser, "settings", task.settings_schema)
    _add_schema_to_group(parser, "inputs", task.input_schema)
    if not no_output:
        _add_schema_to_group(
            parser, "outputs", task.output_schema, {'required': False}, is_output=True)
    return parser


def register_parser(type_desc, parser):
    """Registers a parser for atomic values."""
    _parsers[type_desc] = parser


_parsers = {
    'integer': int,
    'data_frame': DataFrameRef,
    'float': float,
    'date': lambda s: dt.datetime.strptime(s, '%Y-%m-%d').date(),
    'dict': JsonRef,
    'selection': lambda x: x
}


# TODO: clean this up
def _add_schema_to_group(parser, group_name, schema, overrides=None, is_output=False):
    group = parser.add_argument_group(group_name)
    overrides = overrides or {}

    for (name, field_schema) in schema.items():

        if field_schema.get('type') == 'list':
            overrides['required'] = overrides.get('required', True) and field_schema.get(
                'required', True)
            group.add_argument(
                "--{}".format(name),
                nargs='+',
                **_get_arguments(field_schema['schema'], overrides))
        elif field_schema.get('type') == 'data_frame' and is_output:
            group.add_argument(
                "--{}".format(name),
                **_get_arguments(field_schema, overrides),
                default=DataFrameRef("{}.csv".format(name)))
        elif field_schema.get('type') == 'dict' and is_output:
            group.add_argument(
                "--{}".format(name),
                **_get_arguments(field_schema, overrides),
                default=JsonRef("{}.json".format(name)))
        elif field_schema.get('type') == 'dict':
            group.add_argument("--{}".format(name), type=JsonRef, metavar='<json>')
        elif field_schema.get('type') == 'selection':
            overrides['required'] = overrides.get('required', True) and field_schema.get(
                'required', True)
            group.add_argument(
                "--{}".format(name), type=selection.parse, metavar='<json>')
        else:
            group.add_argument("--{}".format(name),
                               **_get_arguments(field_schema, overrides))

    return group


def _get_arguments(schema, overrides=None):
    arguments = {}
    if 'type' in schema:
        if schema['type'] in _parsers:
            arguments['type'] = _parsers[schema['type']]
            arguments['metavar'] = "<{}>".format(schema['type'])
        elif schema['type'] == 'boolean':
            arguments['action'] = 'store_true'
    arguments['required'] = schema.get('required', False)
    if 'allowed' in schema:
        arguments['choices'] = schema['allowed']
    for (key, value) in (overrides or {}).items():
        arguments[key] = value
    return arguments



"""Contains a simple combinator language for predicates."""
import json
from abc import ABC, abstractmethod


def to_dict(selection):
    """Convert selection to an object that is built from Python native types.
    """
    if hasattr(selection, 'to_dict'):
        return selection.to_dict()
    return selection


def to_sql(selection):
    """
    Convert selection to a SQL condition (i.e., something that may be used in a
    where clause).
    """
    if hasattr(selection, 'to_sql'):
        return selection.to_sql()
    return str(selection)


def parse(_s):
    """Parse a selection. Currently not implemented.

    :param _s: The string to parse.
    """
    raise NotImplementedError()


class Selection(ABC):
    @abstractmethod
    def to_sql(self):
        pass

    @abstractmethod
    def to_dict(self):
        pass

    def __repr__(self):
        return json.dumps(self.to_dict())


class List(Selection):
    def __init__(self, *elts):
        self._elts = elts

    def to_sql(self):
        return '({})'.format(','.join(to_sql(e) for e in self._elts))

    def to_dict(self):
        return list(self._elts)

    def __eq__(self, other):
        # pylint: disable=protected-access
        if isinstance(other, List):
            return other._elts == self._elts
        return False

    def __repr__(self):
        return repr(self._elts)


class BinaryOperator(Selection):
    op = None

    def __init__(self, left, right):
        self._left = left
        self._right = right

    @abstractmethod
    def to_sql(self):
        pass

    def to_dict(self):
        return {
            'left': to_dict(self._left),
            'operation': self.op,
            'right': to_dict(self._right)
        }

    def __eq__(self, other):
        # pylint: disable=protected-access
        if self.__class__ == other.__class__:
            return other._left == self._left and \
                other._right == self._right and \
                other.op == self.op
        return False

    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__, self._left, self._right)


class Contains(BinaryOperator):
    """Case sensitive substring match."""
    op = 'contains'

    def to_sql(self):
        # Need double % to distinguish from prinft interpolation prefix.
        return '\"{}\" like \'%%{}%%\''.format(to_sql(self._left), to_sql(self._right))


class And(BinaryOperator):

    op = 'and'

    def to_sql(self):
        return '({}) and ({})'.format(to_sql(self._left), to_sql(self._right))


class Or(BinaryOperator):
    op = 'or'

    def to_sql(self):
        return '({}) or ({})'.format(to_sql(self._left), to_sql(self._right))


class In(BinaryOperator):
    """List containment."""
    op = 'in'

    def __init__(self, left, right):
        super().__init__(left, right if isinstance(right, List) else List(*right))

    def to_sql(self):
        return '\"{}\" in {}'.format(to_sql(self._left), to_sql(self._right))

    def to_dict(self):
        return {
            'left': to_dict(self._left),
            'operation': self.op,
            'right': to_dict(self._right)
        }

    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__, self._left, self._right)


class IContains(BinaryOperator):
    """Case insensitive substring match."""
    op = 'icontains'

    def to_sql(self):
        # Need double % to distinguish from prinft interpolation prefix.
        return '\"{}\" ilike \'%%{}%%\''.format(to_sql(self._left), to_sql(self._right))


class SQL(Selection):
    """SQL literal."""

    def __init__(self, sql):
        self._sql = sql

    def to_sql(self):
        return self._sql

    def to_dict(self):
        return {"sql": self.to_sql()}

    def __eq__(self, other):
        # pylint: disable=protected-access
        if not isinstance(other, self.__class__):
            return False
        return other._sql == self._sql



import io
import logging
import os
from abc import ABC, abstractmethod
from contextlib import contextmanager

import boto3

from balsam.store.refs import Encoding

log = logging.getLogger(__name__)


class AbstractRefStore(ABC):
    """
    Allows reading and writing of data via references.  Allows variation in two
    dimensions: the type of the data (e.g, a pandas data frame, or a native
    dictionary), and the storage location (e.g, an S3 bucket, or a local file
    system).
    """

    @abstractmethod
    @contextmanager
    def read(self, path, encoding):
        pass

    @abstractmethod
    @contextmanager
    def write(self, path, encoding):
        pass


class LocalRefStore(AbstractRefStore):
    def __init__(self, base_path=None):
        self._base_path = base_path

    @contextmanager
    def read(self, path, encoding=Encoding.STRING):
        """Provide a read only reference to the contents of the file at path.

        :param path: The location of the file to read.

        :returns: A read only stream (a file handle).
        """
        mode = 'rb' if encoding == Encoding.BINARY else 'r'
        with open(self._expand_path(path), mode) as handle:
            yield handle

    @contextmanager
    def write(self, path, encoding=Encoding.STRING):
        """Provide a write only reference to the file at path.

        :param path: The location of the file to read.

        :returns: A read only stream (a file handle).
        """
        full_path = self._expand_path(path)
        self._ensure_dir_exists(full_path)
        mode = 'wb' if encoding == Encoding.BINARY else 'w'
        with open(self._expand_path(path), mode) as handle:
            yield handle

    def _expand_path(self, path):
        if self._base_path is None:
            return path
        return os.path.join(self._base_path, path)

    @staticmethod
    def _ensure_dir_exists(full_path):
        try:
            os.makedirs(os.path.dirname(full_path))
        except OSError:
            pass

    def __repr__(self):
        return '<{}({})>'.format(self.__class__.__name__, self._base_path)


class InMemoryRefStore(AbstractRefStore):
    """Reference store that keeps data in memory."""

    def __init__(self):
        self._data = {}

    @contextmanager
    def write(self, path, encoding=Encoding.STRING):
        resource = io.StringIO()
        yield resource
        resource.seek(0)
        self._data[path] = resource.read()

    @contextmanager
    def read(self, path, encoding=Encoding.STRING):
        yield io.StringIO(self._data[path])

    def read_raw(self, path):
        return self._data.get(path)

    def write_raw(self, path, contents):
        self._data[path] = contents

    def __repr__(self):
        return '<{}()>'.format(self.__class__.__name__)


class RemoteRefStoreS3(AbstractRefStore):
    """Reference store that puts data on S3."""

    # Reuse the client over multiple puts and gets
    _client = None

    def __init__(self, bucket, prefix, config=None):
        """Instantiate a ReferenceStore that puts its data on S3.

        :param bucket: The bucket to store data in.
        :param prefix: The prefix within that bucket, under which to store data.
        :param config: Additional configuration passed to the boto3 client
            constructor.  This might include, `aws_key_id` and
            `aws_secret_access_key`.

        :returns: Instance of a subclass of :class:`~AbstractRefStore`.
        """

        self._bucket = bucket
        self._prefix = prefix
        self._config = self._extract_config(config)

    @classmethod
    def _extract_config(cls, config):
        if not config:
            return {}
        result = {}
        for key in ['aws_access_key_id', 'aws_secret_access_key']:
            if key in config:
                result[key] = config[key]
        return result

    @property
    def s3(self):
        if self._client is None:
            log.info('Creating S3 client with access key ID: %s',
                     self._config.get('aws_access_key_id'))
            self._client = boto3.client('s3', **self._config)
        return self._client

    @contextmanager
    def read(self, path, encoding=Encoding.STRING):
        key = self._expand_path(path)
        response = self.s3.get_object(Bucket=self._bucket, Key=key)
        yield response['Body']

    @contextmanager
    def write(self, path, encoding=Encoding.STRING):
        resource = ChimearicIO()
        yield resource
        resource.seek(0)
        key = self._expand_path(path)
        self.s3.put_object(Body=resource, Bucket=self._bucket, Key=key)

    def _expand_path(self, path):
        return os.path.join(self._prefix, path)

    def __repr__(self):
        return '<{}({}:{})>'.format(self.__class__.__name__, self._bucket, self._prefix)


class ChimearicIO(io.BytesIO):
    """BytesIO wrapper that additionally takes care of encoding strings before
    writing them to the bytes buffer.
    """

    def write(self, s):
        if isinstance(s, str):
            super().write(s.encode('utf8'))
        else:
            super().write(s)



import cerberus
import pandas as pd
import numpy as np

from balsam.selection import Selection

_data_frame = cerberus.TypeDefinition('data_frame', (pd.DataFrame, ), ())

cerberus.Validator.types_mapping['data_frame'] = _data_frame

_array = cerberus.TypeDefinition('array', (np.ndarray, ), ())

cerberus.Validator.types_mapping['array'] = _array

_selection = cerberus.TypeDefinition('selection', (Selection, ), ())

cerberus.Validator.types_mapping['selection'] = _selection



from enum import IntEnum
from abc import ABC, abstractmethod

from balsam.schema import check


class Check(IntEnum):
    """
    Enumeration to indicate to a task which of the settings, inputs and outputs
    to check.
    """
    NONE = 0x00
    SETTINGS = 0x01
    INPUT = 0x02
    OUTPUT = 0x04
    ALL = 0x07


class AbstractTask(ABC):
    """
    Interface definition for a task that takes two types of inputs: settings,
    and actual inputs, and returns output values.  A class implementing this
    interface is required to specify the schemas of all of its legal input and
    output values, i.e., to describe its (runtime) type as a function.
    """

    @property
    @abstractmethod
    def settings_schema(self):
        pass

    @property
    @abstractmethod
    def input_schema(self):
        pass

    @property
    @abstractmethod
    def output_schema(self):
        pass

    @abstractmethod
    def run(self, settings, inputs):
        pass

    def __call__(self, settings, inputs, checks=Check.SETTINGS | Check.INPUT):
        if Check.SETTINGS & checks:
            check(self.settings_schema, settings, {'task': self.__class__.__name__})
        if Check.INPUT & checks:
            check(self.input_schema, inputs, {'task': self.__class__.__name__})
        output = self.run(settings, inputs)
        if Check.OUTPUT & checks:
            check(self.output_schema, output, {'task': self.__class__.__name__})
        return output



# pylint: disable=wildcard-import, unused-wildcard-import
from .refs import *
from .store import *

all = [
    'DataFrameRef', 'ArrayRef', 'JsonRef', 'create_ref', 'LocalRefStore',
    'InMemoryRefstore', 'RemoteRefStoreS3', 'Encoding'
]



# pylint: disable=wildcard-import, unused-wildcard-import
from balsam.selection.base import *
from balsam.selection.from_dict import *

__all__ = [
    'to_sql', 'List', 'Contains', 'And', 'Or', 'In', 'Selection', 'to_dict', 'parse',
    'from_dict', 'SQL', 'IContains'
]



import os
from enum import Enum
import logging
import json
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd

from balsam.store.encoder import CustomEncoder

log = logging.getLogger(__name__)


def create_ref(location, value):
    """Factory for :class:`~AbstractRef` instances.  The type of the instance
    depends on the data.

    :param location: Location the reference points to.  Does not include the
        extension.  This location is relative to some :class:`~RefStore`.
    :param value: The data for which the result should be a reference.
    """
    if isinstance(value, pd.DataFrame):
        return DataFrameRef(location + '.csv')
    if isinstance(value, np.ndarray):
        return ArrayRef(location + '.csv')
    if isinstance(value, (dict, str, int, float)):
        return JsonRef(location + '.json')
    raise ValueError("No reference for value of type '{}'".format(type(value)))


class Encoding(Enum):
    BINARY = 'b'
    STRING = ''


class AbstractRef(ABC):
    """
    Abstract base class for references, i.e., store-independent pointers to some
    data file.
    """

    def __init__(self, path):
        self.path = path

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.path == other.path

    @abstractmethod
    def parse(self, store):
        pass

    @abstractmethod
    def write(self, value, store):
        pass

    def read(self, store):
        return self.parse(store)

    def __hash__(self):
        return hash(self.path)

    def __repr__(self):
        return "<{}({})>".format(self.__class__.__name__, self.path)


class LoggedRefMixin:
    """Mixin for logging write and parse calls."""

    def parse(self, store):
        log.info('Reading %s from %s', self.path, store)

    def write(self, _value, store):
        log.info('Writing %s to %s', self.path, store)


class DataFrameRef(LoggedRefMixin, AbstractRef):
    """Reference to a pandas data frame."""

    def parse(self, store):
        """Reads the reference from the store.

        :param store: The store to read from.
        :returns: A data frame.
        """
        super().parse(store)
        sep = self._compute_seperator(self.path)

        with store.read(self.path) as handle:
            df = pd.read_csv(handle, sep=sep, header='infer', parse_dates=True)
            if 'date' in df.columns:
                log.info('Converting column to date')
                df['date'] = pd.to_datetime(
                    df['date'], infer_datetime_format=True).dt.date

            return df

    def write(self, value, store):
        """Writes the dataframe to the store.

        :param value: The data frame value to write.
        :param store: The store to write to.
        :returns: None
        """
        super().write(value, store)
        with store.write(self.path) as handle:
            value.to_csv(handle, index=False)

    @staticmethod
    def _compute_seperator(path):
        (_, ext) = os.path.splitext(path)
        if ext in ['.tsv']:
            return '\t'
        return ','


class JsonRef(LoggedRefMixin, AbstractRef):
    """Reference to a JSON file."""

    def parse(self, store):
        """Reads the reference from the store.

        :param stream: A store.
        :returns: A Python object.
        """
        super().parse(store)
        with store.read(self.path) as handle:
            value = json.load(handle)
            return self._read_arrays(value, store)

    def write(self, value, store):
        """Writes the JSON encoded value to buf.

        :param buf: The buf the write value to.
        :param value: The value to serialize.
        :returns: None
        """
        super().write(value, store)
        expanded = self._store_arrays(value, store)
        with store.write(self.path) as handle:
            json.dump(expanded, handle, cls=CustomEncoder)

    def _store_arrays(self, value, store):
        if isinstance(value, dict):
            expanded = value.copy()
            for (name, item) in value.items():
                if isinstance(item, pd.DataFrame):
                    (file_name, file_path) = self._create_array_file_name_and_path(
                        name, 'csv')
                    expanded[name] = file_name
                    DataFrameRef(file_path).write(item, store)
                if isinstance(item, np.ndarray):
                    (file_name, file_path) = self._create_array_file_name_and_path(
                        name, 'csv')
                    expanded[name] = file_name
                    ArrayRef(file_path).write(item, store)
            return expanded
        return value

    def _create_array_file_name_and_path(self, name, ext):
        file_name = '{name}.{ext}'.format(name=name, ext=ext)
        file_dir = os.path.dirname(self.path)
        return (file_name, os.path.join(file_dir, file_name))

    def _read_arrays(self, value, store):
        if isinstance(value, dict):
            for (name, item) in value.items():
                if self._is_csv(item):
                    value[name] = ArrayRef(item).read(store)
        return value

    def _is_csv(self, name):
        return hasattr(name, 'split') and name.split('.')[-1] == 'csv'


class ArrayRef(LoggedRefMixin, AbstractRef):
    """Reference to a file containing a numpy array."""

    def parse(self, store):
        super().parse(store)
        with store.read(self.path) as handle:
            return np.loadtxt(handle, delimiter=',')

    def write(self, value, store):
        super().write(value, store)
        with store.write(self.path) as handle:
            np.savetxt(handle, value, delimiter=',', fmt='%s')


class FileRef(LoggedRefMixin, AbstractRef):
    """Reference to either a text file or a binary file."""

    def __init__(self, path, encoding=Encoding.STRING):
        if not isinstance(encoding, Encoding):
            raise ValueError('encoding needs to be of type Encoding')
        super().__init__(path)
        self._encoding = encoding

    def parse(self, store):
        super().parse(store)
        with store.read(self.path, self._encoding) as handle:
            return handle.read()

    def write(self, value, store):
        super().write(value, store)
        with store.write(self.path, self._encoding) as handle:
            return handle.write(value)
